<?php

namespace Drupal\commerce_recurly\PluginForm;

use Drupal\commerce\Response\NeedsRedirectException;
use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm as BasePaymentOffsiteForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;

/**
 * Form for checking out via the RecurlyJs gateway.
 *
 * This is not loaded when the user selects a payment method.
 * Since it is associated with offsite payment gateways the
 * form is loaded up after the user has review their order
 * and clicked to "submit payment." At that point the hit the
 * 'payment-redirect-form', which is this form.
 *
 * @package Drupal\commerce_recurly\PluginForm
 */
class RecurlyJsPaymentForm extends BasePaymentOffsiteForm {

  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = $this->entity;

    /** @var \Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayInterface $payment_gateway_plugin */
    $payment_gateway_plugin = $payment->getPaymentGateway()->getPlugin();

    $configuration = $payment_gateway_plugin->getConfiguration();

    $currency_formatter = \Drupal::service('commerce_price.currency_formatter');
    $order_price = $payment->getOrder()->getTotalPrice();

    $data = [
      'public_key' => $configuration['public_key'],
      'return_url' => $form['#return_url'],
      'cancel_url' => $form['#cancel_url'],
      'order_total' => $currency_formatter->format($order_price->getNumber(), $order_price->getCurrencyCode()),
    ];

    $form['notice'] = [
      '#markup' => t("<h1>Authorize Your Purchase</h1><p>PayPal login and order review will happen in a new window. If your browser tries to block any popups on this page, please allow the popup to display.</p>"),
    ];

    // Recurly token
    $form['recurly_token_type'] = [
      '#attributes' => [
        'id' => 'field_recurly_token_type',
      ],
      '#type' => 'hidden',
      '#required' => TRUE,
    ];

    $form['recurly_token_id'] = [
      '#attributes' => [
        'id' => 'field_recurly_token_id',
      ],
      '#type' => 'hidden',
      '#required' => TRUE,
    ];

    // Form actions
    $form['actions'] = ['#type' => 'actions'];

    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Submit'),
    ];

    $form['actions']['cancel'] = [
      '#type' => 'link',
      '#title' => $this->t('Cancel'),
      '#url' => Url::fromUri($form['#cancel_url']),
    ];


    $form['#attached']['library'][] = 'commerce_recurly/recurlyjs';
    $form['#attached']['library'][] = 'commerce_recurly/checkout';
    $form['#attached']['drupalSettings']['commerce_recurly'] = json_encode($data);

    return $this->buildRedirectForm($form, $form_state, $form['#return_url'], $data, self::REDIRECT_POST);
  }

  protected function buildRedirectForm(array $form, FormStateInterface $form_state, $redirect_url, array $data, $redirect_method = self::REDIRECT_GET) {
    if ($redirect_method == self::REDIRECT_POST) {
      $form['#attached']['library'][] = 'commerce_payment/offsite_redirect';
      $form['#process'][] = [get_class($this), 'processRedirectForm'];
      $form['#redirect_url'] = $redirect_url;

      foreach ($data as $key => $value) {
        $form[$key] = [
          '#type' => 'hidden',
          '#value' => $value,
          '#parents' => [$key],
        ];
      }
    }
    else {
      $redirect_url = Url::fromUri($redirect_url, ['absolute' => TRUE, 'query' => $data])->toString();
      throw new NeedsRedirectException($redirect_url);
    }

    return $form;
  }

  public static function processRedirectForm(array $form, FormStateInterface $form_state, array &$complete_form) {
    $complete_form['#action'] = $form['#redirect_url'];
    $complete_form['#attributes']['class'][] = 'payment-redirect-form';

    return $form;
  }

  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state); // TODO: Change the autogenerated stub
  }

}
